/* 
These classes provide two interfaces to the QUB data file format.
CQDFFileAcq for QUBIO ( very fast but very limited functionality ) 
CQDSFFile for general purpose ( i.e. GEM ) ( much faster than QUB but not the direct write QUBIO needs ) 

Not as many features as the QUB version, but useful when you want : 
	simple interface 
	speed 
	no python 
	less in memory buffering ( esp. during acquisition ) 
	no crashing 
	
todo - document this, unless a new file format is on the way.	
*/

const int MAX_DEPTH = 256;		// max depth of node that can be backtracked using GoParent()

const int COUNT_OFFSET = 8;		// offset within node to the count field
//used node types
const BYTE QTP_EMPTY = 0;		
const BYTE QTP_STRING = 3;		
const BYTE QTP_SHORT = 7;		
const BYTE QTP_LONG = 9;		
const BYTE QTP_FLOAT = 12;
const BYTE QTP_DOUBLE = 13;		
//unused node types
const BYTE QTP_TYPE_UNKNOWN=1;
const BYTE QTP_TYPE_POINTER=2;
const BYTE QTP_TYPE_UCHAR=4;
const BYTE QTP_TYPE_CHAR=5;
const BYTE QTP_TYPE_USHORT=6;
const BYTE QTP_TYPE_UINT=8;
const BYTE QTP_TYPE_ULONG=10;
const BYTE QTP_TYPE_LONG=11;
const BYTE QTP_TYPE_LDOUBLE=14;

//---- contents of one node of q?f file.  32 bytes + name + data 
class CQFNode { 
	public : 
		BYTE m_aFlags[3];		// PRELOAD : 2 ; DATA_IN_NODE : 4; MATRIX : 8;
		BYTE m_btType;			// Node Type - QTP_EMPTY, QTP_STRING, ... 
		int m_iSize;			// size of each data item, for example recording 8 channels of shorts = 2*8 = 16
		int m_iCount;			// number of data items.  m_iSize*m_iCount->total data bytes.
		int m_iDataPos;			// file offset to data 
		int m_iChildPos;		// file offset to first child node of this node
		int m_iSiblingPos;		// file offset to next sibling node of this node
		BYTE m_aReserved[7];	
		CString m_cName;		// name of node.
		
		int m_iPos;
		
	public : 
		//-- fill in all class members
		void Set( bool bPreload, BYTE btType, int iSize, int iCount, int iDataPos, 
					int iChildPos, int iSiblingPos, const char * pchName );
					
		//-- fill in all class members from node on disk ( excluding node data ) 
		bool LoadFromFile( FILE * h, int iPos=-1 );
		
		//-- Save all class members to file node at specified pos, or current pos if -1
		//   Optionally set m_iDataPos to location for node data immediately following this node.
		bool SaveToFile( FILE * h, int iPos=-1, bool bSetDataPos=false );

		//-- Load contents of data block.  iLen contains desired read size. 
		//   *piLen receives loaded byte count on exit.
		bool DataBlock( FILE * h, void * pData, int iLen, int iOffset=0, int * piLen=NULL );
		
		//-- Write to the file the 32 byte node info, not including data and name
		bool WriteNode( FILE * h, int iPos );
		
		//-- Load contents of data block as a string
		CString DataString( FILE * h );
	};
	
//---- interface to file 
class CQFFile { 
	protected :
		CString m_cFile;

		int m_aiParent[MAX_DEPTH];		// track parent node list 
		int m_iDepth;
		
	public :
		FILE * m_hFile;

		//-- Constructor - Sets default datatype that will be used, does little else.
		CQFFile( );
		virtual ~CQFFile( );
		//-- Creates a file with 12 byte leader and root node with pointer to first child.
		// Root node is left as 'Current' node.  Returns position of next node (file len)
		int Create( const char * pchFileName, const char * pchRootNodeName );		
		//-- opens the specified file, checks the 12 byte signature 
		bool Open( const char * pchFileName );
		//-- Has file been opened yet.
		bool Opened();
		//-- Close file.
		void Close();

		//---- The 'current' node - used when navigating tree
		CQFNode m_oCurrent;
		CString CurrentString( );		// get string value from current node
		double CurrentDouble( );		// get 1 double value from current node
		int CurrentInt();				// get 1 short or int value from current node 
		BYTE * GetDataBufferToDelete();	// get node data.  The returned buffer must be delete[]'d by caller.

		//----- Maintain list of 'free space' - areas in file no longer used 
		void FreeSpaceAdd( int iPos, int iLen );
		void FreeSpaceFind( int iLen );

		//---- Change the 'current' node
		bool GoRoot();				// load root node to current
		bool GoLastSibling();		// load last sibling of current node to current
		bool GoSibling();			// load next sibling to current, false=no sibling
		bool GoSibling(const char * pchName);	// load next sibling with specified name, false=no sibling
		bool GoChild();
		bool GoParent();		
		bool GoChild( const char * pchName, bool bAdd=false, BYTE btType=QTP_EMPTY, int iSize=0, int iCount=0 );
									// go to child.  if no child and bAdd, add as new child.
		bool AddNode( const char * pchName, bool bChild, BYTE btType, int iSize, int iCount, const void * pData=NULL, int iDataSize=0, bool bPreload=true );
		bool AddStringNode( const char * pchName, bool bChild, CString * pcData, bool bIncludeZero=false  );
		bool AddStringNode( const char * pchName, bool bChild, const char * pchData, bool bIncludeZero=false );
		bool DeleteNode();
		bool ReplaceNodeData( int iPos, int iSize, int iCount, void * pData );

	protected : //-- remember current pos to later be restored 
		int m_aiSaveParent[MAX_DEPTH];		// Restore parent chain list 
		int m_iSaveDepth;					
		int m_iSavePos;						// position to reload current node from 
		void SaveCurrent();
		void RestoreCurrent();
		
	};

//======================================================================================
//-- General interface to combined qdf/qsf file format.
#ifndef QUBIO
	class CQDSFFile : public CQFFile { 
		protected : 
							
		public : 
			//-- Write the "DataFile".Sampling,Scaling,ADChannelCount,ADDataType,ADDDataSize nodes.
			// Current node in and out should be root.
			bool WriteHeader( double dSampling, double dScaling, int iChannels, int iDataSize, int iDataType, bool bSetPos=true );

			//-- Write the "Datafile"."Comment" node.  Current node in and out should be root.
			bool WriteComment( const char * pch, bool bSetPos=false );

			//-- Write the Root."Segments" collection.  Current node in and out should be Root.
			// aiSegOffsets must be allocated for Segments.Count()+1 and will be filled in.
			// adScaling should have [iChannels] elements, already filled in.
			bool WriteSegments( int iChannels, double * adScaling, CItemArray * paSegments, int * aiSegOffsets, double dFileScale );
			
			//-- Write the "Channels" collection.  Current node in and out should be SessionFile.
			bool WriteChannels( CDataConfigure * pConfig );

			//-- Write the "Lists" node.  Current node in and out should be SessionFile.
			bool WriteLists( CItemArray * paLists, int * aiSegOffsets );

			//-- Write the "Idealization" node.  Current node in and out should be SessionFile.
			bool WriteIDL( CItemArray * paChannels, CItemArray * paSegments, int * aiSegOffsets, double dSampling );
			
		};
#else
	//======================================================================================
	//-- Acquisition Interface to qdf data file.  This class optimizes for data written 
	// sequentially (recording).  AddData calls should not succeed any node creation without
	// an intermediate AddDataNode() since both rely on file length checks.
	class CQDFFileAcq : public CQFFile  { 
		public : 
			int m_iSamplesWritten;			// samples written to current segment 
		protected : 
			int m_iSamplesPos;				// Position to write the SamplesCount
			int m_iDataPos;					// pos of data being recorded ( add Sample count offset for current pos )

			int m_iDataType;				// QTP_DOUBLE QTP_FLOAT QTP_SHORT
			int m_iDataSize;				// 8          4         2 
			
			bool m_bCurrentSaved;			// 'Current' node was saved for later restore.
			
			BYTE * m_pBuffer;
			int m_iBufferSize;
			bool AllocateBuffer( int iSize );

		public :
			CQDFFileAcq( int iType = QTP_DOUBLE );
			virtual ~CQDFFileAcq();

			//-- Preset data type to double, float, or short.  Also sets datasize.
			void SetDataType(int iDataType);

			//-- helper func - create the list for the data notes if it D.N.E.
			// Root -> SessionFile -> Lists -> list name will be added to the tree
			bool FindOrAddList( const char * pchListName ) ;
			
			//-- Creates a qdf file with 5 header nodes and 'segments' node, ready for data.
			bool CreateQdf( const char * pchFileName, double dSampling, int iChannels );	
			//-- start a new data node	
			bool AddDataNode( int iChannels );
			//-- add to prior data node.  Convert from double to float or short if required.
			bool AddData( double * pData, int iAnaChannels, int * aiData, int iDigChannels, BYTE * aiDigMap,
						int iSamples, bool lUpdateNode );
			//-- add digital data to prior node 
			bool AddDigoutNode( int iSamples, unsigned int * pData );
	
			//-- adds a channel node and its three child nodes : Name,Units,Scaling.
			bool AddChannel( const char * pchName, const char * pchUnits, double dScaling, double dGain );

			//-- Manage data notes list.
			CItemArray m_aDataNotes;		// array of notes not yet written
			//-- add a note to the memory array 
			bool AddNote( int iPos, const char * pchNote, const char * pchComment );	
			//-- write the memory array to disk.
			bool WriteDataNotes( );
			
			bool WriteComment( const char * pchText );
		};	
				
	//-- Interface to data file list item.
	class CQFDataNote : public CItem { 
		public : 
			int m_iStart, m_iEnd;
			CString m_cNote;
			CString m_cComment;
			CQFDataNote( int iPos, const char * pchNote, const char * pchComment );
		};
#endif	

//======================================================================================
//-- Interface to Model file 
class CQMFFile : public CQFFile { 
	public : 

	};
			
